@@ -11,15 +11,15 @@ use self::{
         coding::{CloseCode, Control as OpCtl, Data as OpData, OpCode},
         Frame, FrameCodec,
     },
-    message::{IncompleteMessage, IncompleteMessageType},
+    message::{IncompleteMessage, MessageType},
 };
 use crate::{
-    error::{Error, ProtocolError, Result},
-    util::NonBlockingResult,
+    error::{CapacityError, Error, ProtocolError, Result},
+    protocol::frame::Utf8Bytes,
 };
 use log::*;
 use std::{
-    io::{ErrorKind as IoErrorKind, Read, Write},
+    io::{self, Read, Write},
     mem::replace,
 };
 
@@ -33,11 +33,28 @@ pub enum Role {
 }
 
 /// The configuration for WebSocket connection.
+///
+/// # Example
+/// ```
+/// # use tungstenite::protocol::WebSocketConfig;;
+/// let conf = WebSocketConfig::default()
+///     .read_buffer_size(256 * 1024)
+///     .write_buffer_size(256 * 1024);
+/// ```
 #[derive(Debug, Clone, Copy)]
+#[non_exhaustive]
 pub struct WebSocketConfig {
-    /// Does nothing, instead use `max_write_buffer_size`.
-    #[deprecated]
-    pub max_send_queue: Option<usize>,
+    /// Read buffer capacity. This buffer is eagerly allocated and used for receiving
+    /// messages.
+    ///
+    /// For high read load scenarios a larger buffer, e.g. 128 KiB, improves performance.
+    ///
+    /// For scenarios where you expect a lot of connections and don't need high read load
+    /// performance a smaller buffer, e.g. 4 KiB, would be appropriate to lower total
+    /// memory usage.
+    ///
+    /// The default value is 128 KiB.
+    pub read_buffer_size: usize,
     /// The target minimum size of the write buffer to reach before writing the data
     /// to the underlying stream.
     /// The default value is 128 KiB.
@@ -58,11 +75,11 @@ pub struct WebSocketConfig {
     /// Note: Should always be at least [`write_buffer_size + 1 message`](Self::write_buffer_size)
     /// and probably a little more depending on error handling strategy.
     pub max_write_buffer_size: usize,
-    /// The maximum size of a message. `None` means no size limit. The default value is 64 MiB
+    /// The maximum size of an incoming message. `None` means no size limit. The default value is 64 MiB
     /// which should be reasonably big for all normal use-cases but small enough to prevent
     /// memory eating by a malicious user.
     pub max_message_size: Option<usize>,
-    /// The maximum size of a single message frame. `None` means no size limit. The limit is for
+    /// The maximum size of a single incoming message frame. `None` means no size limit. The limit is for
     /// frame payload NOT including the frame header. The default value is 16 MiB which should
     /// be reasonably big for all normal use-cases but small enough to prevent memory eating
     /// by a malicious user.
@@ -77,9 +94,8 @@ pub struct WebSocketConfig {
 
 impl Default for WebSocketConfig {
     fn default() -> Self {
-        #[allow(deprecated)]
-        WebSocketConfig {
-            max_send_queue: None,
+        Self {
+            read_buffer_size: 128 * 1024,
             write_buffer_size: 128 * 1024,
             max_write_buffer_size: usize::MAX,
             max_message_size: Some(64 << 20),
@@ -90,6 +106,42 @@ impl Default for WebSocketConfig {
 }
 
 impl WebSocketConfig {
+    /// Set [`Self::read_buffer_size`].
+    pub fn read_buffer_size(mut self, read_buffer_size: usize) -> Self {
+        self.read_buffer_size = read_buffer_size;
+        self
+    }
+
+    /// Set [`Self::write_buffer_size`].
+    pub fn write_buffer_size(mut self, write_buffer_size: usize) -> Self {
+        self.write_buffer_size = write_buffer_size;
+        self
+    }
+
+    /// Set [`Self::max_write_buffer_size`].
+    pub fn max_write_buffer_size(mut self, max_write_buffer_size: usize) -> Self {
+        self.max_write_buffer_size = max_write_buffer_size;
+        self
+    }
+
+    /// Set [`Self::max_message_size`].
+    pub fn max_message_size(mut self, max_message_size: Option<usize>) -> Self {
+        self.max_message_size = max_message_size;
+        self
+    }
+
+    /// Set [`Self::max_frame_size`].
+    pub fn max_frame_size(mut self, max_frame_size: Option<usize>) -> Self {
+        self.max_frame_size = max_frame_size;
+        self
+    }
+
+    /// Set [`Self::accept_unmasked_frames`].
+    pub fn accept_unmasked_frames(mut self, accept_unmasked_frames: bool) -> Self {
+        self.accept_unmasked_frames = accept_unmasked_frames;
+        self
+    }
+
     /// Panic if values are invalid.
     pub(crate) fn assert_valid(&self) {
         assert!(
@@ -147,6 +199,11 @@ impl<Stream> WebSocket<Stream> {
         }
     }
 
+    /// Consumes the `WebSocket` and returns the underlying stream.
+    pub fn into_inner(self) -> Stream {
+        self.socket
+    }
+
     /// Returns a shared reference to the inner stream.
     pub fn get_ref(&self) -> &Stream {
         &self.socket
@@ -161,7 +218,7 @@ impl<Stream> WebSocket<Stream> {
     /// # Panics
     /// Panics if config is invalid e.g. `max_write_buffer_size <= write_buffer_size`.
     pub fn set_config(&mut self, set_func: impl FnOnce(&mut WebSocketConfig)) {
-        self.context.set_config(set_func)
+        self.context.set_config(set_func);
     }
 
     /// Read the configuration.
@@ -313,6 +370,9 @@ pub struct WebSocketContext {
     incomplete: Option<IncompleteMessage>,
     /// Send in addition to regular messages E.g. "pong" or "close".
     additional_send: Option<Frame>,
+    /// True indicates there is an additional message (like a pong)
+    /// that failed to flush previously and we should try again.
+    unflushed_additional: bool,
     /// The configuration for the websocket session.
     config: WebSocketConfig,
 }
@@ -323,7 +383,8 @@ impl WebSocketContext {
     /// # Panics
     /// Panics if config is invalid e.g. `max_write_buffer_size <= write_buffer_size`.
     pub fn new(role: Role, config: Option<WebSocketConfig>) -> Self {
-        Self::_new(role, FrameCodec::new(), config.unwrap_or_default())
+        let conf = config.unwrap_or_default();
+        Self::_new(role, FrameCodec::new(conf.read_buffer_size), conf)
     }
 
     /// Create a WebSocket context that manages an post-handshake stream.
@@ -331,7 +392,8 @@ impl WebSocketContext {
     /// # Panics
     /// Panics if config is invalid e.g. `max_write_buffer_size <= write_buffer_size`.
     pub fn from_partially_read(part: Vec<u8>, role: Role, config: Option<WebSocketConfig>) -> Self {
-        Self::_new(role, FrameCodec::from_partially_read(part), config.unwrap_or_default())
+        let conf = config.unwrap_or_default();
+        Self::_new(role, FrameCodec::from_partially_read(part, conf.read_buffer_size), conf)
     }
 
     fn _new(role: Role, mut frame: FrameCodec, config: WebSocketConfig) -> Self {
@@ -344,6 +406,7 @@ impl WebSocketContext {
             state: WebSocketState::Active,
             incomplete: None,
             additional_send: None,
+            unflushed_additional: false,
             config,
         }
     }
@@ -391,10 +454,16 @@ impl WebSocketContext {
         self.state.check_not_terminated()?;
 
         loop {
-            if self.additional_send.is_some() {
+            if self.additional_send.is_some() || self.unflushed_additional {
                 // Since we may get ping or close, we need to reply to the messages even during read.
-                // Thus we flush but ignore its blocking.
-                self.flush(stream).no_block()?;
+                match self.flush(stream) {
+                    Ok(_) => {}
+                    Err(Error::Io(err)) if err.kind() == io::ErrorKind::WouldBlock => {
+                        // If blocked continue reading, but try again later
+                        self.unflushed_additional = true;
+                    }
+                    Err(err) => return Err(err),
+                }
             } else if self.role == Role::Server && !self.state.can_read() {
                 self.state = WebSocketState::Terminated;
                 return Err(Error::ConnectionClosed);
@@ -403,7 +472,7 @@ impl WebSocketContext {
             // If we get here, either write blocks or we have nothing to write.
             // Thus if read blocks, just let it return WouldBlock.
             if let Some(message) = self.read_message_frame(stream)? {
-                trace!("Received message {}", message);
+                trace!("Received message {message}");
                 return Ok(message);
             }
         }
@@ -432,7 +501,7 @@ impl WebSocketContext {
         }
 
         let frame = match message {
-            Message::Text(data) => Frame::message(data.into(), OpCode::Data(OpData::Text), true),
+            Message::Text(data) => Frame::message(data, OpCode::Data(OpData::Text), true),
             Message::Binary(data) => Frame::message(data, OpCode::Data(OpData::Binary), true),
             Message::Ping(data) => Frame::ping(data),
             Message::Pong(data) => {
@@ -462,7 +531,9 @@ impl WebSocketContext {
     {
         self._write(stream, None)?;
         self.frame.write_out_buffer(stream)?;
-        Ok(stream.flush()?)
+        stream.flush()?;
+        self.unflushed_additional = false;
+        Ok(())
     }
 
     /// Writes any data in the out_buffer, `additional_send` and given `data`.
@@ -484,18 +555,22 @@ impl WebSocketContext {
         let should_flush = if let Some(msg) = self.additional_send.take() {
             trace!("Sending pong/close");
             match self.buffer_frame(stream, msg) {
-                Err(Error::WriteBufferFull(Message::Frame(msg))) => {
+                Err(Error::WriteBufferFull(msg)) => {
                     // if an system message would exceed the buffer put it back in
                     // `additional_send` for retry. Otherwise returning this error
                     // may not make sense to the user, e.g. calling `flush`.
-                    self.set_additional(msg);
-                    false
+                    if let Message::Frame(msg) = *msg {
+                        self.set_additional(msg);
+                        false
+                    } else {
+                        unreachable!()
+                    }
                 }
                 Err(err) => return Err(err),
                 Ok(_) => true,
             }
         } else {
-            false
+            self.unflushed_additional
         };
 
         // If we're closing and there is nothing to send anymore, we should close the connection.
@@ -532,138 +607,116 @@ impl WebSocketContext {
     }
 
     /// Try to decode one message frame. May return None.
-    fn read_message_frame<Stream>(&mut self, stream: &mut Stream) -> Result<Option<Message>>
-    where
-        Stream: Read + Write,
-    {
-        if let Some(mut frame) = self
+    fn read_message_frame(&mut self, stream: &mut impl Read) -> Result<Option<Message>> {
+        let frame = match self
             .frame
-            .read_frame(stream, self.config.max_frame_size)
+            .read_frame(
+                stream,
+                self.config.max_frame_size,
+                matches!(self.role, Role::Server),
+                self.config.accept_unmasked_frames,
+            )
             .check_connection_reset(self.state)?
         {
-            if !self.state.can_read() {
-                return Err(Error::Protocol(ProtocolError::ReceivedAfterClosing));
-            }
-            // MUST be 0 unless an extension is negotiated that defines meanings
-            // for non-zero values.  If a nonzero value is received and none of
-            // the negotiated extensions defines the meaning of such a nonzero
-            // value, the receiving endpoint MUST _Fail the WebSocket
-            // Connection_.
-            {
-                let hdr = frame.header();
-                if hdr.rsv1 || hdr.rsv2 || hdr.rsv3 {
-                    return Err(Error::Protocol(ProtocolError::NonZeroReservedBits));
-                }
+            None => {
+                // Connection closed by peer
+                return match replace(&mut self.state, WebSocketState::Terminated) {
+                    WebSocketState::ClosedByPeer | WebSocketState::CloseAcknowledged => {
+                        Err(Error::ConnectionClosed)
+                    }
+                    _ => Err(Error::Protocol(ProtocolError::ResetWithoutClosingHandshake)),
+                };
             }
+            Some(frame) => frame,
+        };
 
-            match self.role {
-                Role::Server => {
-                    if frame.is_masked() {
-                        // A server MUST remove masking for data frames received from a client
-                        // as described in Section 5.3. (RFC 6455)
-                        frame.apply_mask()
-                    } else if !self.config.accept_unmasked_frames {
-                        // The server MUST close the connection upon receiving a
-                        // frame that is not masked. (RFC 6455)
-                        // The only exception here is if the user explicitly accepts given
-                        // stream by setting WebSocketConfig.accept_unmasked_frames to true
-                        return Err(Error::Protocol(ProtocolError::UnmaskedFrameFromClient));
-                    }
-                }
-                Role::Client => {
-                    if frame.is_masked() {
-                        // A client MUST close a connection if it detects a masked frame. (RFC 6455)
-                        return Err(Error::Protocol(ProtocolError::MaskedFrameFromServer));
-                    }
-                }
-            }
-
-            match frame.header().opcode {
-                OpCode::Control(ctl) => {
-                    match ctl {
-                        // All control frames MUST have a payload length of 125 bytes or less
-                        // and MUST NOT be fragmented. (RFC 6455)
-                        _ if !frame.header().is_final => {
-                            Err(Error::Protocol(ProtocolError::FragmentedControlFrame))
-                        }
-                        _ if frame.payload().len() > 125 => {
-                            Err(Error::Protocol(ProtocolError::ControlFrameTooBig))
-                        }
-                        OpCtl::Close => Ok(self.do_close(frame.into_close()?).map(Message::Close)),
-                        OpCtl::Reserved(i) => {
-                            Err(Error::Protocol(ProtocolError::UnknownControlFrameType(i)))
-                        }
-                        OpCtl::Ping => {
-                            let data = frame.into_data();
-                            // No ping processing after we sent a close frame.
-                            if self.state.is_active() {
-                                self.set_additional(Frame::pong(data.clone()));
-                            }
-                            Ok(Some(Message::Ping(data)))
-                        }
-                        OpCtl::Pong => Ok(Some(Message::Pong(frame.into_data()))),
-                    }
-                }
-
-                OpCode::Data(data) => {
-                    let fin = frame.header().is_final;
-                    match data {
-                        OpData::Continue => {
-                            if let Some(ref mut msg) = self.incomplete {
-                                msg.extend(frame.into_data(), self.config.max_message_size)?;
-                            } else {
-                                return Err(Error::Protocol(
-                                    ProtocolError::UnexpectedContinueFrame,
-                                ));
-                            }
-                            if fin {
-                                Ok(Some(self.incomplete.take().unwrap().complete()?))
-                            } else {
-                                Ok(None)
-                            }
-                        }
-                        c if self.incomplete.is_some() => {
-                            Err(Error::Protocol(ProtocolError::ExpectedFragment(c)))
-                        }
-                        OpData::Text | OpData::Binary => {
-                            let msg = {
-                                let message_type = match data {
-                                    OpData::Text => IncompleteMessageType::Text,
-                                    OpData::Binary => IncompleteMessageType::Binary,
-                                    _ => panic!("Bug: message is not text nor binary"),
-                                };
-                                let mut m = IncompleteMessage::new(message_type);
-                                m.extend(frame.into_data(), self.config.max_message_size)?;
-                                m
-                            };
-                            if fin {
-                                Ok(Some(msg.complete()?))
-                            } else {
-                                self.incomplete = Some(msg);
-                                Ok(None)
-                            }
-                        }
-                        OpData::Reserved(i) => {
-                            Err(Error::Protocol(ProtocolError::UnknownDataFrameType(i)))
-                        }
-                    }
-                }
-            } // match opcode
-        } else {
-            // Connection closed by peer
-            match replace(&mut self.state, WebSocketState::Terminated) {
-                WebSocketState::ClosedByPeer | WebSocketState::CloseAcknowledged => {
-                    Err(Error::ConnectionClosed)
-                }
-                _ => Err(Error::Protocol(ProtocolError::ResetWithoutClosingHandshake)),
+        if !self.state.can_read() {
+            return Err(Error::Protocol(ProtocolError::ReceivedAfterClosing));
+        }
+        // MUST be 0 unless an extension is negotiated that defines meanings
+        // for non-zero values.  If a nonzero value is received and none of
+        // the negotiated extensions defines the meaning of such a nonzero
+        // value, the receiving endpoint MUST _Fail the WebSocket
+        // Connection_.
+        {
+            let hdr = frame.header();
+            if hdr.rsv1 || hdr.rsv2 || hdr.rsv3 {
+                return Err(Error::Protocol(ProtocolError::NonZeroReservedBits));
             }
         }
+
+        if self.role == Role::Client && frame.is_masked() {
+            // A client MUST close a connection if it detects a masked frame. (RFC 6455)
+            return Err(Error::Protocol(ProtocolError::MaskedFrameFromServer));
+        }
+
+        match frame.header().opcode {
+            OpCode::Control(ctl) => {
+                match ctl {
+                    // All control frames MUST have a payload length of 125 bytes or less
+                    // and MUST NOT be fragmented. (RFC 6455)
+                    _ if !frame.header().is_final => {
+                        Err(Error::Protocol(ProtocolError::FragmentedControlFrame))
+                    }
+                    _ if frame.payload().len() > 125 => {
+                        Err(Error::Protocol(ProtocolError::ControlFrameTooBig))
+                    }
+                    OpCtl::Close => Ok(self.do_close(frame.into_close()?).map(Message::Close)),
+                    OpCtl::Reserved(i) => {
+                        Err(Error::Protocol(ProtocolError::UnknownControlFrameType(i)))
+                    }
+                    OpCtl::Ping => {
+                        let data = frame.into_payload();
+                        // No ping processing after we sent a close frame.
+                        if self.state.is_active() {
+                            self.set_additional(Frame::pong(data.clone()));
+                        }
+                        Ok(Some(Message::Ping(data)))
+                    }
+                    OpCtl::Pong => Ok(Some(Message::Pong(frame.into_payload()))),
+                }
+            }
+
+            OpCode::Data(data) => {
+                let fin = frame.header().is_final;
+
+                let payload = match (data, self.incomplete.as_mut()) {
+                    (OpData::Continue, None) => Err(ProtocolError::UnexpectedContinueFrame),
+                    (OpData::Continue, Some(incomplete)) => {
+                        incomplete.extend(frame.into_payload(), self.config.max_message_size)?;
+                        Ok(None)
+                    }
+                    (_, Some(_)) => Err(ProtocolError::ExpectedFragment(data)),
+                    (OpData::Text, _) => Ok(Some((frame.into_payload(), MessageType::Text))),
+                    (OpData::Binary, _) => Ok(Some((frame.into_payload(), MessageType::Binary))),
+                    (OpData::Reserved(i), _) => Err(ProtocolError::UnknownDataFrameType(i)),
+                }?;
+
+                match (payload, fin) {
+                    (None, true) => Ok(Some(self.incomplete.take().unwrap().complete()?)),
+                    (None, false) => Ok(None),
+                    (Some((payload, t)), true) => {
+                        check_max_size(payload.len(), self.config.max_message_size)?;
+                        match t {
+                            MessageType::Text => Ok(Some(Message::Text(payload.try_into()?))),
+                            MessageType::Binary => Ok(Some(Message::Binary(payload))),
+                        }
+                    }
+                    (Some((payload, t)), false) => {
+                        let mut incomplete = IncompleteMessage::new(t);
+                        incomplete.extend(payload, self.config.max_message_size)?;
+                        self.incomplete = Some(incomplete);
+                        Ok(None)
+                    }
+                }
+            }
+        } // match opcode
     }
 
     /// Received a close frame. Tells if we need to return a close frame to the user.
     #[allow(clippy::option_option)]
-    fn do_close<'t>(&mut self, close: Option<CloseFrame<'t>>) -> Option<Option<CloseFrame<'t>>> {
-        debug!("Received close frame: {:?}", close);
+    fn do_close(&mut self, close: Option<CloseFrame>) -> Option<Option<CloseFrame>> {
+        debug!("Received close frame: {close:?}");
         match self.state {
             WebSocketState::Active => {
                 self.state = WebSocketState::ClosedByPeer;
@@ -672,7 +725,7 @@ impl WebSocketContext {
                     if !frame.code.is_allowed() {
                         CloseFrame {
                             code: CloseCode::Protocol,
-                            reason: "Protocol violation".into(),
+                            reason: Utf8Bytes::from_static("Protocol violation"),
                         }
                     } else {
                         frame
@@ -680,7 +733,7 @@ impl WebSocketContext {
                 });
 
                 let reply = Frame::close(close.clone());
-                debug!("Replying to close with {:?}", reply);
+                debug!("Replying to close with {reply:?}");
                 self.set_additional(reply);
 
                 Some(close)
@@ -712,7 +765,7 @@ impl WebSocketContext {
             }
         }
 
-        trace!("Sending frame: {:?}", frame);
+        trace!("Sending frame: {frame:?}");
         self.frame.buffer_frame(stream, frame).check_connection_reset(self.state)
     }
 
@@ -728,6 +781,15 @@ impl WebSocketContext {
     }
 }
 
+fn check_max_size(size: usize, max_size: Option<usize>) -> crate::Result<()> {
+    if let Some(max_size) = max_size {
+        if size > max_size {
+            return Err(Error::Capacity(CapacityError::MessageTooLong { size, max_size }));
+        }
+    }
+    Ok(())
+}
+
 /// The current connection state.
 #[derive(Debug, PartialEq, Eq, Clone, Copy)]
 enum WebSocketState {
@@ -774,7 +836,7 @@ impl<T> CheckConnectionReset for Result<T> {
     fn check_connection_reset(self, state: WebSocketState) -> Self {
         match self {
             Err(Error::Io(io_error)) => Err({
-                if !state.can_read() && io_error.kind() == IoErrorKind::ConnectionReset {
+                if !state.can_read() && io_error.kind() == io::ErrorKind::ConnectionReset {
                     Error::ConnectionClosed
                 } else {
                     Error::Io(io_error)
@@ -817,10 +879,10 @@ mod tests {
             0x03,
         ]);
         let mut socket = WebSocket::from_raw_socket(WriteMoc(incoming), Role::Client, None);
-        assert_eq!(socket.read().unwrap(), Message::Ping(vec![1, 2]));
-        assert_eq!(socket.read().unwrap(), Message::Pong(vec![3]));
+        assert_eq!(socket.read().unwrap(), Message::Ping(vec![1, 2].into()));
+        assert_eq!(socket.read().unwrap(), Message::Pong(vec![3].into()));
         assert_eq!(socket.read().unwrap(), Message::Text("Hello, World!".into()));
-        assert_eq!(socket.read().unwrap(), Message::Binary(vec![0x01, 0x02, 0x03]));
+        assert_eq!(socket.read().unwrap(), Message::Binary(vec![0x01, 0x02, 0x03].into()));
     }
 
     #[test]
